import {
    AfterViewInit,
    Component,
    ElementRef,
    OnDestroy,
    OnInit,
    ViewChild
} from '@angular/core';

import { ActivatedRoute } from '@angular/router';

import {
    MatPaginator,
    MatSort,
    MatTableDataSource
} from '@angular/material';

import { BehaviorSubject } from 'rxjs/BehaviorSubject';

import {
    debounceTime,
    delay,
    distinctUntilChanged,
    startWith,
    tap
} from 'rxjs/operators';

import { merge } from 'rxjs/observable/merge';
import { fromEvent } from 'rxjs/observable/fromEvent';

import { CommentsService } from '../services/comments.service';
import { ImagesService } from '../services/images.service';
import { SearchDataSource } from '../services/search.datasource';


@Component({
    selector: 'search',
    templateUrl: './search.component.html',
    styleUrls: [ './search.component.css' ]
})
export class SearchComponent implements OnInit, OnDestroy, AfterViewInit {
    private badgesSubject = new BehaviorSubject<any[]>([]);
    private emotesSubject = new BehaviorSubject<any[]>([]);

    dataSource: SearchDataSource;

    displayedColumns = [
        'video_title',
        'video_recorded_at',
        'offset',
        'commenter_display_name',
        'message_body'
    ];

    @ViewChild(MatPaginator) paginator: MatPaginator;
    @ViewChild(MatSort) sort: MatSort;
    @ViewChild('input1') input1: ElementRef;
    @ViewChild('input2') input2: ElementRef;

    public badges$ = this.badgesSubject.asObservable();
    public emotes$ = this.emotesSubject.asObservable();

    constructor(private route: ActivatedRoute,
                private commentsService: CommentsService,
                private imagesService: ImagesService) { }

    ngOnInit() {
        this.dataSource = new SearchDataSource(this.commentsService);
        this.dataSource.loadComments('', '', 'video_recorded_at', 'desc', 0, 20);
        this.imagesService.getBadges().subscribe(
            (data: any[]) => this.badgesSubject.next(data)
        );
        this.imagesService.getEmotes().subscribe(
            (data: any[]) => this.emotesSubject.next(data)
        );
    }

    ngOnDestroy() {
        this.badgesSubject.complete();
        this.emotesSubject.complete();
    }

    ngAfterViewInit() {
        this.sort.sortChange.subscribe(() => this.paginator.pageIndex = 0);

        fromEvent(this.input1.nativeElement, 'keyup')
            .pipe(
                debounceTime(600),
                distinctUntilChanged(),
                tap(() => {
                    this.paginator.pageIndex = 0;
                    this.loadCommentsPage();
                })
            )
            .subscribe();

        fromEvent(this.input2.nativeElement, 'keyup')
            .pipe(
                debounceTime(600),
                distinctUntilChanged(),
                tap(() => {
                    this.paginator.pageIndex = 0;
                    this.loadCommentsPage();
                })
            )
            .subscribe();

        merge(this.sort.sortChange, this.paginator.page)
            .pipe(
                tap(() => this.loadCommentsPage())
            )
            .subscribe();
    }

    loadCommentsPage() {
        this.dataSource.loadComments(
            this.input1.nativeElement.value,
            this.input2.nativeElement.value,
            this.sort.active,
            this.sort.direction,
            this.paginator.pageIndex,
            this.paginator.pageSize);
    }
}
